//  This file is part of ff3d - http://www.freefem.org/ff3d
//  Copyright (C) 2001, 2002, 2003 Stphane Del Pino

//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2, or (at your option)
//  any later version.

//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.

//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software Foundation,
//  Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  

//  $Id: Instruction.cpp,v 1.24 2006/12/30 01:32:39 delpinux Exp $

#include <config.h>

#include <cstdlib>
#include <fstream>

#include <Mesh.hpp>
#include <MeshWriter.hpp>

#include <Structured3DMesh.hpp>

#include <Instruction.hpp>
#include <FunctionExpression.hpp>

#include <FunctionExpressionFEM.hpp>
#include <FunctionExpressionValue.hpp>

#include <FEMFunctionBuilder.hpp>

#include <StringExpression.hpp>

#include <Scene.hpp>

#include <MeshExpression.hpp>
#include <Variable.hpp>

#include <Information.hpp>

#include <VTKDriver.hpp>

#include <FEMFunction.hpp>
#include <ErrorHandler.hpp>

#include <ScalarFunctionWriterMedit.hpp>
#include <ScalarFunctionWriterRaw.hpp>
#include <ScalarFunctionWriterVTK.hpp>

#include <FieldFunctionWriterMedit.hpp>
#include <FieldFunctionWriterVTK.hpp>

void InstructionExec::execute()
{
  __command->execute();
  std::string command = __command->value();
  ffout(2) << "executing: " << command << '\n';
#ifdef ALLOW_EXEC
  int result = system(command.c_str());
  if (result == -1) {
    throw ErrorHandler(__FILE__,__LINE__,
		       "could not execute: "+command,
		       ErrorHandler::normal);
  }
#else // ALLOW_EXEC
  fferr(2) << "warning: this version was not compiled with 'exec' support.\n";
#endif // ALLOW_EXEC
}

InstructionExec::InstructionExec(ReferenceCounting<StringExpression> command)
  : Instruction(Instruction::exec),
    __command(command)
{
  ;
}

InstructionExec::InstructionExec(const InstructionExec& I)
  : Instruction(I),
    __command(I.__command)
{
  ;
}

InstructionExec:: ~InstructionExec()
{
  ;
}


void InstructionCat::execute()
{
  __filename->execute();

  const std::string filename = __filename->value();
  std::ifstream fin(filename.c_str());

  if (not fin) {
    throw ErrorHandler(__FILE__,__LINE__,
		       "could not open file \""+filename+"\"",
		       ErrorHandler::normal);
  }
  char c = fin.get();
  while (not(fin.eof())) {
    std::cout << c;
    c = fin.get();
  }
}

InstructionCat::
InstructionCat(ReferenceCounting<StringExpression> filename)
  : Instruction(Instruction::cat),
    __filename(filename)
{
  ;
}

InstructionCat::
InstructionCat(const InstructionCat& I)
  : Instruction(I),
    __filename(I.__filename)
{
  ;
}

InstructionCat::
~InstructionCat()
{
  ;
}


InstructionSave::
InstructionSave(ReferenceCounting<FileDescriptor> descriptor,
		ReferenceCounting<StringExpression> fileName,
		ReferenceCounting<MeshExpression> mesh)
  : Instruction(Instruction::save),
    __fileDescriptor(descriptor),
    __fileName(fileName),
    __mesh(mesh)
{
  ;
}

InstructionSave::InstructionSave(const InstructionSave& I)
  : Instruction(I),
    __fileDescriptor(I.__fileDescriptor),
    __fileName(I.__fileName),
    __mesh(I.__mesh)
{
  ;
}

InstructionSave::~InstructionSave()
{
  ;
}

InstructionSaveMesh::
InstructionSaveMesh(ReferenceCounting<FileDescriptor> descriptor,
		    ReferenceCounting<StringExpression> fileName,
		    ReferenceCounting<MeshExpression> mesh)
  : InstructionSave(descriptor, fileName, mesh)
{
  ;
}

InstructionSaveMesh::
InstructionSaveMesh(const InstructionSaveMesh& I)
  : InstructionSave(I)
{
  ;
}

InstructionSaveMesh::
~InstructionSaveMesh()
{
  ;
}

void InstructionSaveMesh::execute()
{
#warning SHOULD MOVE THIS PIECE OF CODE
  __mesh->execute();
  __fileName->execute();
  Information::instance().setMesh(__mesh->mesh());

  const std::string CR = (*__fileDescriptor).cr();

  switch((*__fileDescriptor).format()) {
  case FileDescriptor::raw: {
    std::string msh = (*__fileName).value();
    throw ErrorHandler(__FILE__,__LINE__,
		       "cannot use raw format to save a mesh!",
		       ErrorHandler::normal);
    break;
  }
  case FileDescriptor::medit: {
    //! plots the Mesh
    std::string msh = (*__fileName).value();

    msh += ".mesh";
    
    std::ofstream fmsh(msh.c_str());
    if (not(fmsh)) {
      throw ErrorHandler(__FILE__,__LINE__,
			 "cannot open file '"+msh+"'",
			 ErrorHandler::normal);
    }

    MeshWriter w(fmsh,(*__mesh).mesh(),CR);

    break;
  }
  default: {
    throw ErrorHandler(__FILE__,__LINE__,
		       "unexpected file type",
		       ErrorHandler::normal);
  }
  }

  //! The mesh is nomore used in this region.
  Information::instance().unsetMesh();
}

InstructionSaveFunction::
InstructionSaveFunction(ReferenceCounting<FileDescriptor> descriptor,
			ReferenceCounting<StringExpression> fileName,
			ReferenceCounting<FunctionExpression> function,
			ReferenceCounting<MeshExpression> mesh)
  : InstructionSave(descriptor, fileName, mesh),
    __function(function)
{
  ;
}

InstructionSaveFunction::
InstructionSaveFunction(const InstructionSaveFunction& I)
  : InstructionSave(I),
    __function(I.__function)
{
  ;
}

InstructionSaveFunction::
~InstructionSaveFunction()
{
  ;
}

void InstructionSaveFunction::execute()
{
  __mesh->execute();
  __function->execute();
  __fileName->execute();
  Information::instance().setMesh(__mesh->mesh());

  const std::string CR = __fileDescriptor->cr();

  ReferenceCounting<ScalarFunctionWriterBase> writer;

  switch(__fileDescriptor->format()) {
  case FileDescriptor::medit: {
    writer = new ScalarFunctionWriterMedit(__function->function(),
					   __mesh->mesh(),
					   __fileName->value(),
					   __fileDescriptor->cr());
    break;
  }
  case FileDescriptor::raw: {
    writer = new ScalarFunctionWriterRaw(__function->function(),
					 __mesh->mesh(),
					 __fileName->value(),
					 __fileDescriptor->cr());
    break;
  }
  case FileDescriptor::vtk: {
    writer = new ScalarFunctionWriterVTK(__function->function(),
					 __mesh->mesh(),
					 __fileName->value(),
					 __fileDescriptor->cr());
    break;
  }
  default: {
    fferr(0) << __FILE__ << ':' << __LINE__ << ": Not implemented\n";
    std::exit(1);
  }
  }

  writer->proceed();

  //! The mesh is nomore used in this region.
  Information::instance().unsetMesh();
}

void
InstructionAffectation<FunctionExpressionFEM, FunctionVariable>::
execute()
{
  __expression->execute();
  ConstReferenceCounting<ScalarFunctionBase> newValue = __expression->function();

  ReferenceCounting<FunctionExpression> f = __variable->expression();

  const FEMFunctionBase& femfunction = dynamic_cast<const FEMFunctionBase&>(*f->function());

  Information::instance().setMesh(femfunction.baseMesh());

  FEMFunctionBuilder builder;
  builder.build(femfunction.discretizationType(),
		femfunction.baseMesh(),
		*newValue);

  (*__variable) = new FunctionExpressionValue(builder.getBuiltFunction(), false);

  Information::instance().unsetMesh();
}


InstructionSaveField::
InstructionSaveField(ReferenceCounting<FileDescriptor> descriptor,
		     ReferenceCounting<StringExpression> fileName,
		     ReferenceCounting<FieldExpression> field,
		     ReferenceCounting<MeshExpression> mesh)
  : InstructionSave(descriptor,fileName,mesh),
    __field(field)
{
  ;
}

InstructionSaveField::
InstructionSaveField(const InstructionSaveField& I)
  : InstructionSave(I),
    __field(I.__field)
{
  ;
}

InstructionSaveField::
~InstructionSaveField()
{
  ;
}

void InstructionSaveField::execute()
{
  __mesh->execute();
  __field->execute();
  __fileName->execute();

  Information::instance().setMesh(__mesh->mesh());

  const std::string CR = __fileDescriptor->cr();

  ReferenceCounting<FieldFunctionWriterBase> writer;

  switch(__fileDescriptor->format()) {
  case FileDescriptor::medit: {
    writer = new FieldFunctionWriterMedit(__field->f1()->function(),
					  __field->f2()->function(),
					  __field->f3()->function(),
					  __mesh->mesh(),
					  __fileName->value(),
					  __fileDescriptor->cr());
    break;
  }
  case FileDescriptor::vtk: {
    writer = new FieldFunctionWriterVTK(__field->f1()->function(),
					__field->f2()->function(),
					__field->f3()->function(),
					__mesh->mesh(),
					__fileName->value(),
					__fileDescriptor->cr());
    break;
  }
  default: {
    throw ErrorHandler(__FILE__,__LINE__,
		       "not implemented",
		       ErrorHandler::unexpected);
  }
  }

  writer->proceed();

  //! The mesh is nomore used in this region.
  Information::instance().unsetMesh();
}

void InstructionPlot::execute()
{
  __mesh->execute();

  Mesh& M = *(__mesh->mesh());
  VTKDriver d;

  if (__f == 0) {
    d.plot(M);
  } else {
    __f->execute();
    ConstReferenceCounting<ScalarFunctionBase> u
      = __f->value()->function();
    d.plot(M, u);
  }
}

InstructionPlot::InstructionPlot(ReferenceCounting<MeshExpression> m,
				 ReferenceCounting<FunctionExpression> f)
  : Instruction(Instruction::plot),
    __mesh(m), 
    __f(f)
{
  ;
}

InstructionPlot::InstructionPlot(const InstructionPlot& I)
  : Instruction(I),
    __mesh(I.__mesh)
{
  ;
}

InstructionPlot::~InstructionPlot()
{
  ;
}

void InstructionUsingScene::execute()
{
  __sceneExpression->execute();
  Information::instance().setScene(__sceneExpression->scene());
}

InstructionUsingScene::InstructionUsingScene(ReferenceCounting<SceneExpression> e)
  : Instruction(Instruction::Using),
    __sceneExpression(e)
{
  ;
}

InstructionUsingScene::InstructionUsingScene(const InstructionUsingScene& I)
  : Instruction(I),
    __sceneExpression(I.__sceneExpression)
{
  ;
}

InstructionUsingScene::~InstructionUsingScene()
{
  ;
}

void InstructionCoarseMesh::execute()
{
  Information::instance().setCoarseMesh(__coarseMesh);
}

