//  This file is part of ff3d - http://www.freefem.org/ff3d
//  Copyright (C) 2001, 2002, 2003 Stphane Del Pino

//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2, or (at your option)
//  any later version.

//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.

//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software Foundation,
//  Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  

//  $Id: MeshWriter.cpp,v 1.11 2006/09/28 20:08:51 delpinux Exp $

#include <Mesh.hpp>

#include <MeshWriter.hpp>

#include <fstream>
#include <Structured3DMesh.hpp>
#include <MeshOfTetrahedra.hpp>
#include <MeshOfHexahedra.hpp>
#include <MeshOfTriangles.hpp>

#include <SurfaceMeshOfTriangles.hpp>
#include <SurfaceMeshOfQuadrangles.hpp>

template <typename MeshType>
void MeshWriter::__saveElements(ConstReferenceCounting<MeshType> m)
{
  typedef typename MeshType::CellType CellType;

  __os << m->numberOfCells() << __cr;

  for (size_t i = 0; i<m->numberOfCells(); ++i) {
    const CellType& c = m->cell(i);

    for (size_t j=0; j<CellType::NumberOfVertices; ++j) {
      __os << m->vertexNumber(c(j))+1 << ' ';
    }

    __os << c.reference() << __cr;
  }
}


void MeshWriter::__proceed()
{
  __os << "MeshVersionFormatted 1" << __cr;
  __os << "Dimension" << __cr;
  __os << 3 << __cr;

  __os << "Vertices " << (*__mesh).numberOfVertices() << __cr;

  for (size_t i=0; i<(*__mesh).numberOfVertices(); ++i) {
    const TinyVector<3>& X = (*__mesh).vertex(i);
    __os << X[0] << ' ' <<  X[1] << ' ' << X[2]
	 << ' ' << (*__mesh).vertex(i).reference() << __cr;
  }

  const Mesh* m = __mesh;
  switch ((*__mesh).type()) {
  case Mesh::cartesianHexahedraMesh: {
    ConstReferenceCounting<Structured3DMesh>
      mesh = dynamic_cast<const Structured3DMesh*>(m);
    __os << "Hexahedra ";
    this->__saveElements(mesh);
    if (mesh->hasSurfaceMesh()) {
      __os << "Quadrilaterals ";
      this->__saveElements(mesh->surfaceMesh());
    }
    break;
  }
  case Mesh::hexahedraMesh: {
    ConstReferenceCounting<MeshOfHexahedra>
      mesh = dynamic_cast<const MeshOfHexahedra*>(m);
    __os << "Hexahedra ";
    this->__saveElements(mesh);
    if (mesh->hasSurfaceMesh()) {
      __os << "Quadrilaterals ";
      this->__saveElements(mesh->surfaceMesh());
    }
    break;
  }
  case Mesh::tetrahedraMesh: {
    ConstReferenceCounting<MeshOfTetrahedra>
      mesh = dynamic_cast<const MeshOfTetrahedra*>(m);
    __os << "Tetrahedra ";
    this->__saveElements(mesh);
    if (mesh->hasSurfaceMesh()) {
      __os << "Triangles ";
      this->__saveElements<SurfaceMeshOfTriangles>(mesh->surfaceMesh());
    }
    break;
  }
  case Mesh::trianglesMesh: {
    ConstReferenceCounting<MeshOfTriangles>
      mesh = dynamic_cast<const MeshOfTriangles*>(m);
    __os << "Triangles ";
    this->__saveElements(mesh);
    break;
  }
  case Mesh::surfaceMeshTriangles: {
    ConstReferenceCounting<SurfaceMeshOfTriangles>
      mesh = dynamic_cast<const SurfaceMeshOfTriangles*>(m);
    __os << "Triangles ";
    this->__saveElements(mesh);
    break;
  }
  case Mesh::surfaceMeshQuadrangles: {
    ConstReferenceCounting<SurfaceMeshOfQuadrangles>
      mesh = dynamic_cast<const SurfaceMeshOfQuadrangles*>(m);
    __os << "Quadrilaterals ";
    this->__saveElements(mesh);
    break;
  }
  default: {
    throw ErrorHandler(__FILE__,__LINE__,
		       "unexpected mesh type",
		       ErrorHandler::unexpected);
  }
  }

  __os << "End" << __cr;
}
