//  This file is part of ff3d - http://www.freefem.org/ff3d
//  Copyright (C) 2001, 2002, 2003 Stphane Del Pino

//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2, or (at your option)
//  any later version.

//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.

//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software Foundation,
//  Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  

//  $Id: ScalarFunctionWriterVTK.cpp,v 1.2 2006/11/11 18:08:46 delpinux Exp $

#include <ScalarFunctionWriterVTK.hpp>

#include <ScalarFunctionBase.hpp>
#include <FEMFunctionBase.hpp>

#include <Mesh.hpp>
#include <Structured3DMesh.hpp>
#include <MeshOfTetrahedra.hpp>

#include <XMLWriter.hpp>

#include <fstream>

template <typename CellType>
struct ScalarFunctionWriterVTK::Traits {};

template <>
struct ScalarFunctionWriterVTK::Traits<Tetrahedron>
{
  enum {
    VTKType = 10
  };
};

template <typename MeshType>
void ScalarFunctionWriterVTK::
__proceed() const
{
  typedef typename MeshType::CellType CellType;

  std::string filename = __fileName;
  filename += ".vtu";

  const MeshType& mesh
    = static_cast<const MeshType&>(*__mesh);

  std::ofstream file(filename.c_str());

  XMLWriter xmlWriter(file);

  xmlWriter.writeHeader();
  {
    XMLTag vtkFile("VTKFile");
    ReferenceCounting<XMLAttribute> attribute
      = new XMLAttribute("type","Unstructured");
    vtkFile.add(attribute);
    xmlWriter.add(vtkFile);
  }

  {
    XMLTag unstructuredGrid("UnstructuredGrid");
    xmlWriter.add(unstructuredGrid);
  }

  {
    XMLTag piece("Piece");
    ReferenceCounting<XMLAttribute>
      nbPoints = new XMLAttribute("NumberOfPoints",
				  stringify(mesh.numberOfVertices()));
    piece.add(nbPoints);
    ReferenceCounting<XMLAttribute>
      nbCells = new XMLAttribute("NumberOfCells",
				 stringify(mesh.numberOfCells()));
    piece.add(nbCells);
    xmlWriter.add(piece);

    {
      XMLTag points("Points");
      xmlWriter.add(points);
   
      {  
	XMLTag dataarray("DataArray");
	ReferenceCounting<XMLAttribute>
	  type = new XMLAttribute("type","Float32");
	dataarray.add(type);
	ReferenceCounting<XMLAttribute>
	  name = new XMLAttribute("Name","Position");
	dataarray.add(name);
	ReferenceCounting<XMLAttribute>
	  nbComponent = new XMLAttribute("NumberOfComponents","3");
	dataarray.add(nbComponent);
	ReferenceCounting<XMLAttribute>
	  format = new XMLAttribute("format","ascii");
	dataarray.add(format);
    
	xmlWriter.add(dataarray);

	for (size_t i=0; i<mesh.numberOfVertices(); ++i) {
	  const Vertex& x = mesh.vertex(i); 
	  xmlWriter.insert(x[0]);
	  xmlWriter.insert(x[1]);
	  xmlWriter.insert(x[2]);
	}
	xmlWriter.insertNewLine();

	xmlWriter.closeTag();
      }
      xmlWriter.closeTag();
    }

    {
      XMLTag cells("Cells");
      xmlWriter.add(cells);

      {
	XMLTag dataarray("DataArray");
	ReferenceCounting<XMLAttribute>
	  type = new XMLAttribute("type","Int32");
	dataarray.add(type);
	ReferenceCounting<XMLAttribute>
	  name = new XMLAttribute("Name","connectivity");
	dataarray.add(name);
	ReferenceCounting<XMLAttribute>
	  nbComponent = new XMLAttribute("NumberOfComponents","1");
	dataarray.add(nbComponent);
	ReferenceCounting<XMLAttribute>
	  format = new XMLAttribute("format","ascii");
	dataarray.add(format);
	xmlWriter.add(dataarray);

	for (size_t i=0;i<mesh.numberOfCells(); ++i) {
	  const CellType& cell = mesh.cell(i);
	  for (size_t j=0; j<CellType::NumberOfVertices; ++j) {
	    xmlWriter.insert(mesh.vertexNumber(cell(j)));
	  }
	}
	xmlWriter.insertNewLine();

	xmlWriter.closeTag();      
      }

      {
	XMLTag dataarray("DataArray");
	ReferenceCounting<XMLAttribute>
	  type = new XMLAttribute("type","Int32");
	dataarray.add(type);
	ReferenceCounting<XMLAttribute>
	  name = new XMLAttribute("Name","offsets");
	dataarray.add(name);
	ReferenceCounting<XMLAttribute>
	  nbComponent = new XMLAttribute("NumberOfComponents","1");
	dataarray.add(nbComponent);
	ReferenceCounting<XMLAttribute>
	  format = new XMLAttribute("format","ascii");
	dataarray.add(format);
	xmlWriter.add(dataarray);
     
	for (size_t i=1; i<=mesh.numberOfCells(); ++i) {
	  xmlWriter.insert(i*CellType::NumberOfVertices);
	}
	xmlWriter.insertNewLine();

	xmlWriter.closeTag();      
      }

      {
	XMLTag dataarray("DataArray");
	ReferenceCounting<XMLAttribute>
	  type = new XMLAttribute("type","UInt8");
	dataarray.add(type);
	ReferenceCounting<XMLAttribute>
	  name = new XMLAttribute("Name","types");
	dataarray.add(name);
	ReferenceCounting<XMLAttribute>
	  nbComponent = new XMLAttribute("NumberOfComponents","1");
	dataarray.add(nbComponent);
	ReferenceCounting<XMLAttribute>
	  format = new XMLAttribute("format","ascii");
	dataarray.add(format);
	xmlWriter.add(dataarray);
     
	for (size_t i=0; i<mesh.numberOfCells(); ++i) {
	  xmlWriter.insert(Traits<CellType>::VTKType);
	}
	xmlWriter.insertNewLine();

	xmlWriter.closeTag();      
      }
    }
    xmlWriter.closeTag();
  }
  {
    XMLTag pointdata("PointData");
    xmlWriter.add(pointdata);

    {  
      XMLTag dataarray("DataArray");
      ReferenceCounting<XMLAttribute>
	type = new XMLAttribute("type","Float32");
      dataarray.add(type);
      ReferenceCounting<XMLAttribute>
	name = new XMLAttribute("Name","Field1");
      dataarray.add(name);
      ReferenceCounting<XMLAttribute>
	nbComponent = new XMLAttribute("NumberOfComponents","1");
      dataarray.add(nbComponent);
      ReferenceCounting<XMLAttribute>
	format = new XMLAttribute("format","ascii");
      dataarray.add(format);
    
      xmlWriter.add(dataarray);

      const ScalarFunctionBase& f = *__function;
      for (size_t i=0; i<mesh.numberOfVertices(); ++i) {
	const Vertex& x = mesh.vertex(i); 
	xmlWriter.insert(f(x));
      }
      
      xmlWriter.insertNewLine();

      xmlWriter.closeTag();
    }

    xmlWriter.closeTag();
  }
}

template <>
void ScalarFunctionWriterVTK::
__proceed<Structured3DMesh>() const
{
  std::string filename = __fileName;
  filename += ".vti";

  const Structured3DMesh& mesh
    = static_cast<const Structured3DMesh&>(*__mesh);

  std::ofstream file(filename.c_str());

  XMLWriter xmlWriter(file);

  xmlWriter.writeHeader();
  {
    XMLTag vtkFile("VTKFile");
    ReferenceCounting<XMLAttribute> attribute
      = new XMLAttribute("type","ImageData");
    vtkFile.add(attribute);
    xmlWriter.add(vtkFile);
  }

  {
    XMLTag rectilinearGrid("ImageData");
    ReferenceCounting<XMLAttribute> wholeExtent
      = new XMLAttribute("WholeExtent",
			   "1 "+stringify(mesh.shape().nx())
			 +" 1 "+stringify(mesh.shape().ny())
			 +" 1 "+stringify(mesh.shape().nz()));
    rectilinearGrid.add(wholeExtent);
    ReferenceCounting<XMLAttribute> origin
      = new XMLAttribute("Origin",
			 stringify(mesh.shape().a(0))+" "+
			 stringify(mesh.shape().a(1))+" "+
			 stringify(mesh.shape().a(2)));
    rectilinearGrid.add(origin);
    ReferenceCounting<XMLAttribute> spacing
      = new XMLAttribute("Spacing",
			 stringify(mesh.shape().hx())+" "+
			 stringify(mesh.shape().hy())+" "+
			 stringify(mesh.shape().hz()));
    rectilinearGrid.add(spacing);
    xmlWriter.add(rectilinearGrid);
  }

  {
    XMLTag piece("Piece");
    ReferenceCounting<XMLAttribute> extent
      = new XMLAttribute("Extent",
			   "1 "+stringify(mesh.shape().nx())
			 +" 1 "+stringify(mesh.shape().ny())
			 +" 1 "+stringify(mesh.shape().nz()));
    piece.add(extent);
    xmlWriter.add(piece);
  }

  {
    XMLTag pointData("PointData");
    XMLAttribute scalars("Scalars",__fileName);
    xmlWriter.add(pointData);
    XMLTag dataArray("DataArray");
    ReferenceCounting<XMLAttribute> type
      = new XMLAttribute("type","Float32");
    dataArray.add(type);
    ReferenceCounting<XMLAttribute> name
      = new XMLAttribute("Name",__fileName);
    dataArray.add(name);
    ReferenceCounting<XMLAttribute> format
      = new XMLAttribute("format","ascii");
    dataArray.add(format);
    xmlWriter.add(dataArray);

    const ScalarFunctionBase& function = *__function;

    switch (function.type()) {
    case ScalarFunctionBase::femfunction: {
      const FEMFunctionBase& fem
	= static_cast<const FEMFunctionBase&>(function);

      if ((fem.discretizationType() == DiscretizationType::LagrangianFEM1)
	  and (fem.baseMesh() == __mesh)) {
	for (size_t k=0; k<mesh.shape().nz(); ++k)
	  for (size_t j=0; j<mesh.shape().ny(); ++j)
	    for (size_t i=0; i<mesh.shape().nx(); ++i) {
	      const size_t pointNumber = mesh.shape()(i,j,k);
	      xmlWriter.insert(fem[pointNumber]);
	    }
	break;
      }
    }
    default: {
      for (size_t k=0; k<mesh.shape().nz(); ++k)
	for (size_t j=0; j<mesh.shape().ny(); ++j)
	  for (size_t i=0; i<mesh.shape().nx(); ++i) {
	    const Vertex& x = mesh.vertex(i,j,k);
	    xmlWriter.insert(function(x));
	  }
    }
    }

    xmlWriter.insertNewLine();

    xmlWriter.closeTag();
    xmlWriter.closeTag();
  }
}

void ScalarFunctionWriterVTK::
proceed() const
{
  switch (__mesh->type()) {
  case Mesh::cartesianHexahedraMesh: {
    return this->__proceed<Structured3DMesh>();
  }
  case Mesh::tetrahedraMesh: {
    return this->__proceed<MeshOfTetrahedra>();
  }
  default: {
    throw ErrorHandler(__FILE__,__LINE__,
		       "not implemented",
		       ErrorHandler::unexpected);
  }
  }
}

ScalarFunctionWriterVTK::
ScalarFunctionWriterVTK(ConstReferenceCounting<ScalarFunctionBase> function,
			ConstReferenceCounting<Mesh> mesh,
			const std::string& fileName,
			const std::string& CR)
  : ScalarFunctionWriterBase(function, mesh, fileName, CR)
{
  ;
}

ScalarFunctionWriterVTK::
~ScalarFunctionWriterVTK()
{
  ;
}
